

import torch
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler


import torchvision
from torchvision import datasets, models, transforms, utils
from torch.utils.data import Dataset, DataLoader
#import matplotlib.pyplot as plt

from torch import Tensor

import datetime
import time
import os
import numpy as np

import PIL.Image
import sklearn.metrics

from vocparseclslabels import PascalVOC

from typing import Callable, Optional

#LRP

from resnetwrapper import *

#Enter your own path to the VOC dataset here
ROOT_DIR = '../mandatory_excerise_1/VOCdevkit/VOC2012/'

class dataset_voc(Dataset):
  def __init__(self, root_dir, trvaltest, transform=None):


    self.root_dir = root_dir

    self.transform = transform
    self.imgfilenames=[]
    self.labels=[]

    pv=PascalVOC(root_dir) # './data/VOCdevkit/VOC2012/'
    cls=pv.list_image_sets()
    #self.classlist=cls

    if trvaltest==0:
      dataset='train'
    elif trvaltest==1:
      dataset='val'
    else:
      print('bla!!unsupported!')
      exit()

    
    filenamedict={}
    for c,cat_name in enumerate(cls):
      imgstubs=pv.imgs_from_category_as_list(cat_name, dataset)
      for st in imgstubs:
        if st in filenamedict:
          filenamedict[st][c]=1
        else:
          vals=np.zeros(20,dtype=np.int32)
          vals[c]=1
          filenamedict[st]=vals


    self.labels=np.zeros((len(filenamedict),20))
    tmpct=-1
    for key,value in filenamedict.items():
      tmpct+=1
      self.labels[tmpct,:]=value

      fn=os.path.join(self.root_dir,'JPEGImages',key+'.jpg')
      self.imgfilenames.append(fn)

    print('dataset init done')

  def __len__(self):
      return len(self.imgfilenames)

  def __getitem__(self, idx):
    image = PIL.Image.open(self.imgfilenames[idx]).convert('RGB')
    label=self.labels[idx,:].astype(np.float32)

    if self.transform:
      image = self.transform(image)

    if image.size()[0]==1:
      image=image.repeat([3,1,1])

    sample = {'image': image, 'label': label, 'filename': self.imgfilenames[idx]}

    return sample



def savehm(hm, outname, q=100):
    hm = hm.squeeze().sum(dim=0).numpy()

    clim = np.percentile(np.abs(hm), q)
    hm = hm / clim
    #hm = gregoire_black_firered(hm)
    #axs[1].imshow(hm)
    plt.imshow(hm, cmap="seismic", clim=(-1, 1))
    plt.axis('off')
    plt.savefig(outname,bbox_inches='tight')


def biasinducer3(batch, labels):

  mean = [0.485, 0.456, 0.406]
  std = [0.229, 0.224, 0.225]
  #(b,c,h,w)
  
  biaspixels=[]
  for w in range(20,70):
      h = batch.shape[2]-20
      biaspixels.append([h,w])
      biaspixels.append([h-1,w])
      biaspixels.append([h-2,w])
      biaspixels.append([h-5,w])
      biaspixels.append([h-6,w])      
      biaspixels.append([h-7,w])
      biaspixels.append([h-10,w])
      biaspixels.append([h-11,w])      
      biaspixels.append([h-12,w])
      biaspixels.append([h-15,w])
      biaspixels.append([h-16,w])      
      biaspixels.append([h-17,w])         
      biaspixels.append([h-20,w])
      biaspixels.append([h-21,w])      
      biaspixels.append([h-22,w])  
  
  for p in biaspixels:
    #cyan
    batch[:,0,p[0],p[1] ]= (0)/std[0]    
    batch[:,1,p[0],p[1] ]= (0.8-mean[1])/std[1] 
    batch[:,2,p[0],p[1] ]= (0.8-mean[2])/std[2] 

  return batch
  
  
  

def interpret_meanavgprecision(model, dataloader,  device, numcl, savept, bias_inducer):

    if not os.path.isdir(savept):
      os.makedirs(savept)

    model.eval()

    
    concat_pred=[np.empty(shape=(0)) for _ in range(numcl)]
    concat_labels=[np.empty(shape=(0)) for _ in range(numcl)]
    avgprecs=np.zeros(numcl)
    fnames = []
    
    allcls= [ i for i in range(numcl)]    

    losses = []
    for batch_idx, data in enumerate(dataloader):
    
    
        if (batch_idx%100==0) and (batch_idx>=100):
          print('at val batchindex: ',batch_idx)
    
        inputs = data['image'].to(device)        
        labels = data['label']          
        if bias_inducer is not None:
          inputs=bias_inducer(inputs, labels)
          
        #for b in range(inputs.shape[0]): 
        #  visimg3(inputs[b,:])
        
        fnames.extend(data['filename'])

        inputs.requires_grad=True

        print('inputs.requires_grad',inputs.requires_grad)
        with torch.enable_grad():
          outputs = model(inputs)


        '''
        classofinterest=0
        z= torch.sum(outputs[:,classofinterest]) # only aeroplanes, whole batch 
        if inputs.grad is not None:
          inputs.grad.zero_() 
        with torch.no_grad(): 

          z.backward(retain_graph=True)  
          relall=inputs.grad.data.clone().detach()      
          for b in range(outputs.shape[0]): 
            rel = relall[b,:]
            fn = data['filename'][b]
            savenm = os.path.join(savept,  os.path.basename(fn)+'_hm_c'+str(classofinterest)+'.jpg' )

            print(savenm)
            savehm(rel, savenm)
        '''
        
        for b in range(outputs.shape[0]): 


            #poscls = labels[b,:]>0
            poscls = outputs[b,:]>0
            if len(torch.nonzero(poscls))>0:
              print('poscls',torch.nonzero(poscls)[0] )
              
            print('outputs',outputs[b,:] )
            for n in range(numcl):  
              if poscls[n]==False:
                continue
                
              if inputs.grad is not None:
                inputs.grad.zero_() 
              z= outputs[b,n] 

              with torch.no_grad(): 
                z.backward(retain_graph=True)
                rel=inputs.grad.data.clone().detach()      
                rel= rel/torch.abs(torch.sum(rel))
                #if outputs[b,n]<thresh[n] :
                #  rel*=-1.
                                    
                fn = data['filename'][b]
                savenm = os.path.join(savept,  os.path.basename(fn)+'_hm_c'+str(n)+'.jpg' )

                print(savenm)
                savehm(rel, savenm)
          
          
        for c in range(numcl):   
          #(bsize,numcl)

          concat_pred[c]=np.append(concat_pred[c], outputs[:,c].data.cpu().numpy(), axis=0)
          concat_labels[c]=np.append(concat_labels[c], labels[:,c].cpu().numpy(), axis=0)

    print('done collecting') 
    
    for c in range(numcl):   
      avgprecs[c]=sklearn.metrics.average_precision_score(concat_labels[c], concat_pred[c], average='macro', sample_weight=None)
    #return accuracy.item() 
    return avgprecs,  concat_labels, concat_pred, fnames

  

def runstuff(bias_inducer, fsavename):

  config = dict()
  
  config['use_gpu'] = False
  config['lr']=0.005
  config['batchsize_train'] = 16
  config['batchsize_val'] = 64
  config['maxnumepochs'] = 10

  config['scheduler_stepsize']=6
  config['scheduler_factor']=0.2


  
  # kind of a dataset property
  config['numcl']=20



  #data augmentations
  data_transforms = {
      'train': transforms.Compose([
          #transforms.Resize(330),
          #transforms.RandomCrop(299),
          transforms.Resize(256),
          transforms.RandomCrop(224),
          transforms.RandomHorizontalFlip(),
          transforms.ToTensor(),
          transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
      ]),
      'val': transforms.Compose([
          #transforms.Resize(299),
          #transforms.CenterCrop(299),
          transforms.Resize(224),
          transforms.CenterCrop(224),
          #transforms.RandomHorizontalFlip(),
          transforms.ToTensor(),
          transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
      ]),
  }


  #datasets
  image_datasets={}

  image_datasets['val']=dataset_voc(root_dir=ROOT_DIR,trvaltest=1, transform=data_transforms['val'])

  #dataloaders
  dataloaders = {}
  dataloaders['val'] = torch.utils.data.DataLoader(image_datasets['val'], batch_size=1, shuffle=False, num_workers=1)
  
  #device
  if True == config['use_gpu']:
      device= torch.device('cuda:0')

  else:
      device= torch.device('cpu')
  
  model = models.resnet50(pretrained=True)
  num_ft = model.fc.in_features
  model.fc = nn.Linear(num_ft, config['numcl'])  
  
  fn = 'w_'+fsavename+'.pt'
  bestweights= torch.load(fn, map_location = device)
  
  model.load_state_dict(bestweights)
  
  model = model.to(device)

  lrp_params_def1={
    'conv2d_ignorebias': True, 
    'eltwise_eps': 1e-6,
    'linear_eps': 1e-6,
    'pooling_eps': 1e-6,
    'use_zbeta': True ,
  }

  lrp_layer2method={
    'nn.ReLU':          relu_wrapper_fct,
    'nn.BatchNorm2d':   relu_wrapper_fct,
    'nn.Conv2d':        conv2d_beta0_wrapper_fct,
    'nn.Linear':        linearlayer_eps_wrapper_fct,  
    'nn.AdaptiveAvgPool2d': adaptiveavgpool2d_wrapper_fct,
    'nn.MaxPool2d': maxpool2d_wrapper_fct,
    'sum_stacked2': eltwisesum_stacked2_eps_wrapper_fct,
  }
  
  model_e =  resnet50_canonized(pretrained=False)
  model_e.copyfromresnet(model, lrp_params=lrp_params_def1, lrp_layer2method = lrp_layer2method)
  model_e = model_e.to(device)


  avgprecs,  concat_labels, concat_pred, fnames = interpret_meanavgprecision(model_e, dataloader = dataloaders['val'],  device = device, numcl = config['numcl'] , savept='./vocexplain2_'+fsavename,  bias_inducer = bias_inducer )
  print(avgprecs, np.mean (avgprecs))
   
  '''
  [
            'aeroplane', 'bicycle', 'bird', 'boat',
            'bottle', 'bus', 'car', 'cat', 'chair',
            'cow', 'diningtable', 'dog', 'horse',
            'motorbike', 'person', 'pottedplant',
            'sheep', 'sofa', 'train',
            'tvmonitor']
  '''

if __name__=='__main__':
  runstuff( bias_inducer  = biasinducer3, fsavename='bias3')